@@ -1,7 +1,7 @@
 /*
  * Functions for transferring data between file descriptors.
  *
- * Copyright 2002-2008, 2010, 2012-2015, 2017, 2021, 2023-2024 Andrew Wood
+ * Copyright 2002-2008, 2010, 2012-2015, 2017, 2021, 2023-2025 Andrew Wood
  *
  * License GPLv3+: GNU GPL version 3 or later; see `docs/COPYING'.
  */
@@ -113,7 +113,7 @@ static int is_data_ready(int fd_in, /*@null@ */ bool *fd_in_ready, int fd_out, /
  * We stop retrying if the time elapsed since this function was entered
  * reaches TRANSFER_READ_TIMEOUT seconds.
  */
-static ssize_t pv__transfer_read_repeated(int fd, void *buf, size_t count)
+static ssize_t pv__transfer_read_repeated(int fd, char *buf, size_t count)
 {
 	struct timespec start_time;
 	ssize_t total_read;
@@ -185,7 +185,7 @@ static ssize_t pv__transfer_read_repeated(int fd, void *buf, size_t count)
  * not guaranteed to succeed for _all_ sizes; we may end up returning 0 if this
  * occurs. (The first write() may return -1 / EINTR if the consumer doesn't
  * read any data before our timeout and the buffer of whatever stdout is is
- * near-full.) (see https://codeberg.org/a-j-wood/pv/pulls/93)
+ * near-full.) (see https://codeberg.org/ivarch/pv/pulls/93)
  *
  * If "sync_after_write" is true, we call fdatasync() after each write() (or
  * fsync() if _POSIX_SYNCHRONIZED_IO is not > 0).
@@ -193,7 +193,7 @@ static ssize_t pv__transfer_read_repeated(int fd, void *buf, size_t count)
  * We stop retrying if the time elapsed since this function was entered
  * reaches TRANSFER_WRITE_TIMEOUT seconds.
  */
-static ssize_t pv__transfer_write_repeated(int fd, void *buf, size_t count, bool sync_after_write)
+static ssize_t pv__transfer_write_repeated(int fd, char *buf, size_t count, bool sync_after_write)
 {
 	struct timespec start_time;
 	ssize_t total_written;
@@ -300,41 +300,63 @@ static ssize_t pv__transfer_write_repeated(int fd, void *buf, size_t count, bool
  * returns 1.
  *
  * At most, the number of bytes read will be the number of bytes remaining
- * in the input buffer.  If state->control.rate_limit is >0, and/or "allowed" is >0,
- * then the maximum number of bytes read will be the number remaining unused
- * in the input buffer or the value of "allowed", whichever is smaller.
+ * in the input buffer, capped to the number of bytes left until
+ * state->control.size is reached if state->control.stop_at_size is true.
+ * If state->control.rate_limit is >0, and/or "max_to_write" is >0, and
+ * splice() is used, then the maximum number of bytes read will be further
+ * capped to the value of "max_to_write", since splice() writes as well as
+ * reads.
  *
- * If splice() was successfully used, sets state->transfer.splice_used to true; if it
- * failed, then state->transfer.splice_failed_fd is updated to the current fd so
- * splice() won't be tried again until the next input file.
+ * If splice() was successfully used, sets state->transfer.splice_used to
+ * true; if it failed, then state->transfer.splice_failed_fd is updated to
+ * the current fd so splice() won't be tried again until the next input
+ * file.
  *
- * Updates state->transfer.read_position by the number of bytes read, unless splice()
- * was used, in which case it does not since there's nothing in the buffer
- * (and it also adds the bytes to state->transfer.written since they've been written
- * to the output).
+ * Updates state->transfer.read_position by the number of bytes read, unless
+ * splice() was used, in which case it does not since there's nothing in the
+ * buffer (and it also adds the bytes to state->transfer.written since
+ * they've been written to the output).
  *
- * On read error, updates state->status.exit_status, and if allowed by
+ * Also increases state->transfer.total_bytes_read by the number of bytes
+ * read, regardless of whether splice() was used, since total_bytes_read is
+ * what it says it is, rather than being a buffer indicator.
+ *
+ * On read error, updates state->status.exit_status, and if max_to_write by
  * state->control.skip_errors, tries to skip past the problem.
  *
  * If the end of the input file is reached or the error is unrecoverable,
  * sets *eof_in to true.  If all data in the buffer has been written at this
  * point, then also sets *eof_out to true.
  */
-static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_out, off_t allowed)
+static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_out, off_t max_to_write)
 {
 	bool do_not_skip_errors;
 	size_t bytes_can_read;
 	off_t amount_to_skip, amount_skipped, orig_offset, skip_offset;
 	ssize_t nread;
-#ifdef HAVE_SPLICE
-	size_t bytes_to_splice;
-#endif				/* HAVE_SPLICE */
 
 	do_not_skip_errors = false;
 	if (0 == state->control.skip_errors)
 		do_not_skip_errors = true;
 
 	bytes_can_read = state->transfer.buffer_size - state->transfer.read_position;
+
+	/*
+	 * Don't read past control.size if stop_at_size is true (issue
+	 * #166).
+	 *
+	 * This isn't workable in line mode.
+	 */
+	if (state->control.stop_at_size && !state->control.linemode) {
+		off_t bytes_remaining_to_read = state->control.size - state->transfer.total_bytes_read;
+		if ((long long) bytes_can_read > (long long) bytes_remaining_to_read) {
+			debug("%lld > (%lld-%lld=%lld): %s", (long long) bytes_can_read,
+			      (long long) (state->control.size), (long long) state->transfer.total_bytes_read,
+			      (long long) bytes_remaining_to_read, "truncating for stop-at-size");
+			bytes_can_read = (size_t) bytes_remaining_to_read;
+		}
+	}
+
 	nread = 0;
 
 #ifdef HAVE_SPLICE
@@ -342,8 +364,10 @@ static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_ou
 	if ((!state->control.linemode) && (!state->control.no_splice)
 	    && (fd != state->transfer.splice_failed_fd)
 	    && (0 == state->transfer.to_write)) {
-		if (state->control.rate_limit > 0 || allowed != 0) {
-			bytes_to_splice = (size_t) allowed;
+		size_t bytes_to_splice;
+
+		if (state->control.rate_limit > 0 || max_to_write != 0) {
+			bytes_to_splice = (size_t) max_to_write;
 		} else {
 			bytes_to_splice = bytes_can_read;
 		}
@@ -365,6 +389,7 @@ static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_ou
 			 */
 		} else if (nread > 0) {
 			state->transfer.written = nread;
+			state->transfer.total_bytes_read += nread;
 #ifdef HAVE_FDATASYNC
 			if (state->control.sync_after_write) {
 				/*
@@ -430,6 +455,8 @@ static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_ou
 #else
 		state->transfer.read_position += nread;
 #endif				/* HAVE_SPLICE */
+		/* Update the counter of all bytes read so far. */
+		state->transfer.total_bytes_read += nread;
 		return 1;
 	}
 
@@ -460,15 +487,7 @@ static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_ou
 	 * reached the end of this file.
 	 */
 	if (do_not_skip_errors) {
-		/*@-compdef@ */
-		pv_error(state, "%s: %s: %s", pv_current_file_name(state), _("read failed"), strerror(errno));
-		/*@+compdef@ */
-		/*
-		 * splint says the storage pointed to by the result of
-		 * pv_current_file_name() is not fully defined.
-		 *
-		 * TODO: investigate and fix the reason for this.
-		 */
+		pv_error("%s: %s: %s", pv_current_file_name(state), _("read failed"), strerror(errno));
 		*eof_in = true;
 		if (state->transfer.write_position >= state->transfer.read_position) {
 			*eof_out = true;
@@ -484,7 +503,7 @@ static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_ou
 
 	if (!state->transfer.read_error_warning_shown) {
 		/*@-compdef@ */
-		pv_error(state, "%s: %s: %s", pv_current_file_name(state), _("warning: read errors detected"),
+		pv_error("%s: %s: %s", pv_current_file_name(state), _("warning: read errors detected"),
 			 strerror(errno));
 		/*@+compdef@ */
 		/* splint - see previous pv_current_file_name() call. */
@@ -500,7 +519,7 @@ static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_ou
 	 */
 	if (0 > orig_offset) {
 		/*@-compdef@ */
-		pv_error(state, "%s: %s: %s", pv_current_file_name(state), _("file is not seekable"), strerror(errno));
+		pv_error("%s: %s: %s", pv_current_file_name(state), _("file is not seekable"), strerror(errno));
 		/*@+compdef@ */
 		/* splint - see previous pv_current_file_name() calls. */
 		*eof_in = true;
@@ -577,8 +596,7 @@ static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_ou
 		 */
 		if (EINVAL != errno) {
 			/*@-compdef@ */
-			pv_error(state,
-				 "%s: %s: %s", pv_current_file_name(state), _("failed to seek past error"),
+			pv_error("%s: %s: %s", pv_current_file_name(state), _("failed to seek past error"),
 				 strerror(errno));
 			/*@+compdef@ */
 			/* splint - see previous pv_current_file_name() calls. */
@@ -596,7 +614,7 @@ static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_ou
 		state->transfer.read_position += amount_skipped;
 		if (state->control.skip_errors < 2) {
 			/*@-compdef@ */
-			pv_error(state, "%s: %s: %ld - %ld (%ld %s)",
+			pv_error("%s: %s: %ld - %ld (%ld %s)",
 				 pv_current_file_name(state),
 				 _("skipped past read error"), (long) orig_offset, (long) skip_offset,
 				 (long) amount_skipped, _("B"));
@@ -637,9 +655,12 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 {
 	ssize_t nwritten;
 	int write_errno;
+	bool all_nulls;
+	off_t output_offset;
+	size_t write_check_position, write_end_position;
 
 	if (NULL == state->transfer.transfer_buffer) {
-		pv_error(state, "%s", _("no transfer buffer allocated"));
+		pv_error("%s", _("no transfer buffer allocated"));
 		state->status.exit_status |= PV_ERROREXIT_MEMORY;
 		*eof_out = true;
 		state->transfer.written = -1;
@@ -653,6 +674,52 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 		nwritten = state->transfer.to_write;
 	} else if (state->transfer.to_write > 0) {
 
+		/*
+		 * In sparse output mode, check whether all of the bytes to
+		 * be written are null, and if so, try to seek the output
+		 * instead of writing the null bytes.
+		 */
+		if (state->control.sparse_output && !state->transfer.output_not_seekable) {
+			write_check_position = state->transfer.write_position;
+			write_end_position = write_check_position + (size_t) (state->transfer.to_write);
+			all_nulls = true;
+			while (all_nulls && write_check_position < write_end_position) {
+				if ('\0' == state->transfer.transfer_buffer[write_check_position]) {
+					write_check_position++;
+				} else {
+					all_nulls = false;
+				}
+			}
+			if (all_nulls) {
+
+				/*
+				 * Use lseek() to move forward in the file,
+				 * and then ftruncate() it to its new size.
+				 *
+				 * If any step fails, mark the output not
+				 * seekable so we stop trying.
+				 */
+
+				/*@+longintegral@ */
+				/*
+				 * splint has trouble with off_t / __off_t, in the lseek() call.
+				 */
+				output_offset =
+				    lseek(state->control.output_fd, (off_t) (state->transfer.to_write), SEEK_CUR);
+				if (output_offset == (off_t) - 1) {
+					debug("%s: %s", "output lseek() failed", strerror(errno));
+					state->transfer.output_not_seekable = true;
+				} else {
+					/* Seek successful - skip write. */
+					debug("%s (%ld) -> %s: %ld", "skipped null writes",
+					      (long) (state->transfer.to_write), "new position", (long) output_offset);
+					nwritten = state->transfer.to_write;
+					goto pv__transfer_write_completed;
+				}
+				/*@-longintegral@ */
+			}
+		}
+
 		/*
 		 * Set an interval timer or an alarm to interrupt the write
 		 * with a signal if the write takes too long, so we can
@@ -681,7 +748,7 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 		      (long) (new_timer.it_value.tv_usec));
 
 		if (0 != setitimer(ITIMER_REAL, &new_timer, NULL)) {
-			pv_error(state, "%s: %s", _("failed to set interval timer"), strerror(errno));
+			pv_error("%s: %s", _("failed to set interval timer"), strerror(errno));
 		}
 
 #else				/* ! HAVE_SETITIMER */
@@ -707,7 +774,7 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 		new_timer.it_value.tv_sec = 0;
 		new_timer.it_value.tv_usec = 0;
 		if (0 != setitimer(ITIMER_REAL, &new_timer, NULL)) {
-			pv_error(state, "%s: %s", _("failed to clear interval timer"), strerror(errno));
+			pv_error("%s: %s", _("failed to clear interval timer"), strerror(errno));
 		}
 
 		/*@+unrecog@ */
@@ -717,29 +784,46 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 #endif				/* HAVE_SETITIMER */
 	}
 
+      pv__transfer_write_completed:
+	/* If lseek() worked for sparse output, it jumps down here. */
+
 	if (nwritten > 0) {
+		bool tracking_lines = false;
+
+		if ((state->control.linemode) && (lineswritten != NULL))
+			tracking_lines = true;
+		else if (state->display.showing_previous_line)
+			tracking_lines = true;
+
 		/*
 		 * Write returned >0 - data successfully written.
 		 */
-		if ((state->control.linemode) && (lineswritten != NULL)) {
+		if (tracking_lines) {
 			char separator;
 			char *ptr;
 			long lines = 0;
 
 			/*
-			 * Line mode - look through what we've just written
-			 * to count how many lines there were.
+			 * Tracking lines - either line mode, or we're
+			 * showing the last line in the display, or both.
+			 * So we need to look through what we've just
+			 * written to either count how many lines there
+			 * were, or get the content of the most recent
+			 * complete line, or both.
 			 */
 
 			/* Allocate buffer to remember line positions. */
-			if (NULL == state->transfer.line_positions) {
+			if (NULL == state->transfer.line_positions && NULL != lineswritten) {
 				state->transfer.line_positions_capacity = MAX_LINE_POSITIONS;
+				/*@-mustfreeonly@ */
 				state->transfer.line_positions =
 				    calloc((size_t) (state->transfer.line_positions_capacity), sizeof(off_t));
 				if (NULL == state->transfer.line_positions) {
-					pv_error(state, "%s: %s", _("line position buffer allocation failed"),
+					pv_error("%s: %s", _("line position buffer allocation failed"),
 						 strerror(errno));
 				}
+				/*@+mustfreeonly@ */
+				/* splint doesn't see we only call calloc() when line_positions is NULL. */
 			}
 
 			if (state->control.null_terminated_lines) {
@@ -752,12 +836,50 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 			for (ptr++;
 			     ptr - (char *) state->transfer.transfer_buffer - state->transfer.write_position <
 			     (size_t) nwritten; ptr++, state->transfer.last_output_position++) {
-				if (*ptr != separator)
+				if (*ptr != separator) {
+					/*
+					 * If we're displaying the previous
+					 * line ("%L"), add to our line
+					 * buffer.
+					 */
+					if (state->display.showing_previous_line
+					    && state->display.next_line_len < PV_SIZEOF_PREVLINE_BUFFER - 1) {
+						state->display.next_line[state->display.next_line_len] = *ptr;
+						state->display.next_line_len++;
+					}
 					continue;
+				}
 
 				/* Separator found - increment line count. */
 				++lines;
 
+				/*
+				 * If we're displaying the previous line
+				 * ("%L"), update the previous-line buffer
+				 * with the line we just completed, and
+				 * start a new one.
+				 */
+				if (state->display.showing_previous_line) {
+					memset(state->display.previous_line, 0, PV_SIZEOF_PREVLINE_BUFFER);
+					if (state->display.next_line_len > PV_SIZEOF_PREVLINE_BUFFER - 1)
+						state->display.next_line_len = PV_SIZEOF_PREVLINE_BUFFER - 1;
+					if (state->display.next_line_len > 0) {
+						memcpy(state->display.previous_line, state->display.next_line,	/* flawfinder: ignore */
+						       state->display.next_line_len);
+						debug("%s: [%s]", "updated previous_line",
+						      state->display.previous_line);
+					}
+					state->display.next_line_len = 0;
+					/*
+					 * flawfinder - next_line_len is
+					 * guaranteed to be less than the
+					 * size of the previous_line buffer
+					 * since we check it just before
+					 * memcpy(), and we ensure that the
+					 * last byte in the buffer is \0.
+					 */
+				}
+
 				if (NULL == state->transfer.line_positions)
 					continue;
 
@@ -777,7 +899,8 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 				}
 			}
 
-			*lineswritten += lines;
+			if (NULL != lineswritten)
+				*lineswritten += lines;
 		}
 
 		state->transfer.write_position += nwritten;
@@ -787,39 +910,39 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 		 * If we're monitoring the output, update our copy of the
 		 * last few bytes we've written.
 		 */
-		if (state->display.component[PV_COMPONENT_OUTPUTBUF].required && (nwritten > 0)) {
-			size_t new_portion_length, old_portion_length;
+		if (state->display.showing_last_written && (nwritten > 0)) {
+			size_t new_portion_size, old_portion_size;
 
-			new_portion_length = (size_t) nwritten;
-			if (new_portion_length > state->display.lastoutput_length)
-				new_portion_length = state->display.lastoutput_length;
+			new_portion_size = (size_t) nwritten;
+			if (new_portion_size > state->display.lastwritten_bytes)
+				new_portion_size = state->display.lastwritten_bytes;
 
-			old_portion_length = state->display.lastoutput_length - new_portion_length;
+			old_portion_size = state->display.lastwritten_bytes - new_portion_size;
 
 			/*
 			 * Make room for the new portion.
 			 */
-			if (old_portion_length > 0) {
-				memmove(state->display.lastoutput_buffer,
-					state->display.lastoutput_buffer + new_portion_length, old_portion_length);
+			if (old_portion_size > 0) {
+				memmove(state->display.lastwritten_buffer,
+					state->display.lastwritten_buffer + new_portion_size, old_portion_size);
 			}
 
 			/*
 			 * Copy the new data in.
 			 */
-			memcpy(state->display.lastoutput_buffer +	/* flawfinder: ignore */
-			       old_portion_length,
-			       state->transfer.transfer_buffer + state->transfer.write_position - new_portion_length,
-			       new_portion_length);
+			memcpy(state->display.lastwritten_buffer +	/* flawfinder: ignore */
+			       old_portion_size,
+			       state->transfer.transfer_buffer + state->transfer.write_position - new_portion_size,
+			       new_portion_size);
 			/*
 			 * flawfinder rationale: calculations above ensure
-			 * that old_portion_length + new_portion_length is
-			 * always <= lastoutput_length, and
-			 * lastoutput_length is guaranteed by
+			 * that old_portion_size + new_portion_size is
+			 * always <= lastwritten_bytes, and
+			 * lastwritten_bytes is guaranteed by
 			 * pv__format_init() to be no more than
-			 * PV_SIZEOF_LASTOUTPUT_BUFFER, which is the size of
-			 * lastoutput_buffer, so the memcpy() will always
-			 * fit into the buffer.
+			 * PV_SIZEOF_LASTWRITTEN_BUFFER, which is the size
+			 * of lastwritten_buffer, so the memcpy() will
+			 * always fit into the buffer.
 			 */
 		}
 
@@ -865,11 +988,12 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 	if (EPIPE == write_errno) {
 		*eof_in = true;
 		*eof_out = true;
-		state->flag.pipe_closed = 1;
+		state->flags.pipe_closed = 1;
+		debug("%s", "SIGPIPE received - setting pipe_closed");
 		return 0;
 	}
 
-	pv_error(state, "%s: %s", _("write failed"), strerror(write_errno));
+	pv_error("%s: %s", _("write failed"), strerror(write_errno));
 	state->status.exit_status |= PV_ERROREXIT_TRANSFER;
 	*eof_out = true;
 	state->transfer.written = -1;
@@ -894,7 +1018,7 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 /*@only@*/
 static char *pv__allocate_aligned_buffer(int outfd, int infd, size_t target_size)
 {
-	char *newptr;
+	void *newptr;
 
 #if defined(HAVE_FPATHCONF) && defined(HAVE_POSIX_MEMALIGN) && defined(_PC_REC_XFER_ALIGN)
 	long input_alignment, output_alignment, min_alignment;
@@ -1008,7 +1132,7 @@ ssize_t pv_transfer(pvstate_t state, int fd, bool *eof_in, bool *eof_out, off_t
 		state->transfer.transfer_buffer =
 		    pv__allocate_aligned_buffer(state->control.output_fd, fd, state->control.target_buffer_size + 32);
 		if (NULL == state->transfer.transfer_buffer) {
-			pv_error(state, "%s: %s", _("buffer allocation failed"), strerror(errno));
+			pv_error("%s: %s", _("buffer allocation failed"), strerror(errno));
 			state->status.exit_status |= PV_ERROREXIT_MEMORY;
 			return -1;
 		}
@@ -1056,8 +1180,10 @@ ssize_t pv_transfer(pvstate_t state, int fd, bool *eof_in, bool *eof_out, off_t
 	if ((state->control.linemode) && (lineswritten != NULL))
 		*lineswritten = 0;
 
-	if ((*eof_in) && (*eof_out))
+	if ((*eof_in) && (*eof_out)) {
+		debug("%s %d: %s", "fd", fd, "early return 0 - EOF in and out");
 		return 0;
+	}
 
 	check_read_fd = -1;
 	check_write_fd = -1;
@@ -1099,15 +1225,16 @@ ssize_t pv_transfer(pvstate_t state, int fd, bool *eof_in, bool *eof_out, off_t
 		/*
 		 * Ignore transient errors by returning 0 immediately.
 		 */
-		if (EINTR == errno)
+		if (EINTR == errno) {
+			debug("%s %d: %s", "fd", fd, "early return 0 - is_data_ready < 0");
 			return 0;
+		}
 
 		/*
 		 * Any other error is a problem and we must report back.
 		 */
 		/*@-compdef@ */
-		pv_error(state, "%s: %s: %d: %s", pv_current_file_name(state), _("select call failed"), n,
-			 strerror(errno));
+		pv_error("%s: %s: %d: %s", pv_current_file_name(state), _("select call failed"), n, strerror(errno));
 		/*@+compdef@ */
 		/* splint - see previous pv_current_file_name() calls. */
 
@@ -1125,8 +1252,12 @@ ssize_t pv_transfer(pvstate_t state, int fd, bool *eof_in, bool *eof_out, off_t
 	 * NB this can update state->transfer.written because of splice().
 	 */
 	if (ready_to_read) {
-		if (pv__transfer_read(state, fd, eof_in, eof_out, allowed) == 0)
+		if (pv__transfer_read(state, fd, eof_in, eof_out, allowed) == 0) {
+			debug("%s %d: %s (%s=%s, %s=%s, %s=%lu)", "fd", fd,
+			      "early return 0 - pv__transfer_read returned 0", "eof_in", eof_in ? "true" : "false",
+			      "eof_out", eof_out ? "true" : "false", "allowed", (unsigned long) allowed);
 			return 0;
+		}
 	}
 
 	/*
@@ -1157,8 +1288,12 @@ ssize_t pv_transfer(pvstate_t state, int fd, bool *eof_in, bool *eof_out, off_t
 	    && (state->transfer.read_position > state->transfer.write_position)
 	    && (state->transfer.to_write > 0)
 	    && (NULL != lineswritten)) {
-		if (pv__transfer_write(state, eof_in, eof_out, lineswritten) == 0)
+		if (pv__transfer_write(state, eof_in, eof_out, lineswritten) == 0) {
+			debug("%s %d: %s (%s=%s, %s=%s, %s=%lu)", "fd", fd,
+			      "early return 0 - pv__transfer_write returned 0", "eof_in", eof_in ? "true" : "false",
+			      "eof_out", eof_out ? "true" : "false", "lineswritten", (unsigned long) lineswritten);
 			return 0;
+		}
 	}
 #ifdef MAXIMISE_BUFFER_FILL
 	/*
@@ -1180,7 +1315,9 @@ ssize_t pv_transfer(pvstate_t state, int fd, bool *eof_in, bool *eof_out, off_t
 	}
 #endif				/* MAXIMISE_BUFFER_FILL */
 
+	if (0 == state->transfer.written) {
+		debug("%s %d: %s", "fd", fd, "end-of-function return 0 - transfer.written is zero");
+	}
+
 	return state->transfer.written;
 }
-
-/* EOF */
